const std = @import("std");

const Optional = struct {
    const OptionEnum = enum {
        Some,
        None,
    };

    pub fn into_option(value: anytype) Option(@typeInfo(@TypeOf(value)).Optional.child) {
        return Option(@typeInfo(@TypeOf(value)).Optional.child).from_optional(value);
    }

    pub fn Option(comptime T: type) type {
        return union(OptionEnum) {
            Some: T,
            None,

            pub inline fn type_() type {
                return T;
            }

            pub inline fn from_optional(value: ?T) Self {
                if (value) |v| {
                    return Self.some(v);
                } else {
                    return Self.none();
                }
            }

            pub fn into_optional_t(self: Self) ?T {
                return switch (self) {
                    .Some => |value| value,
                    .None => null,
                };
            }

            pub inline fn some(value: T) Self {
                return .{ .Some = value };
            }

            pub inline fn none() Self {
                return Option(T).None;
            }

            const Self = @This();

            pub fn is_some(self: Self) bool {
                return switch (self) {
                    .Some => true,
                    .None => false,
                };
            }

            pub fn is_none(self: Self) bool {
                return !self.is_some();
            }

            pub fn unwrap(self: Self) T {
                return self.Some;
            }

            pub fn map(self: Self, comptime U: type, func: fn (T) U) Option(U) {
                return switch (self) {
                    .Some => |value| Option(U).some(func(value)),
                    .None => Option(U).none(),
                };
            }

            pub fn or_option(self: Self, other: Self) Self {
                return switch (self) {
                    .Some => self,
                    .None => other,
                };
            }

            pub fn map_with(self: Self, with: anytype) Option(@typeInfo(@TypeOf(@TypeOf(with).call)).Fn.return_type.?) {
                const Return = Option(@typeInfo(@TypeOf(@TypeOf(with).call)).Fn.return_type.?);
                return switch (self) {
                    .Some => |value| Return.some(with.call(value)),
                    .None => Return.none(),
                };
            }

            pub fn join(self: Self, comptime U: type, func: fn (T) Option(U)) Option(U) {
                return switch (self) {
                    .Some => |value| func(value),
                    .None => Option(U).none(),
                };
            }

            pub fn join_with(self: Self, with: anytype) Option(@typeInfo(@TypeOf(@TypeOf(with).call)).Fn.return_type.?.type_()) {
                const Return = Option(@typeInfo(@TypeOf(@TypeOf(with).call)).Fn.return_type.?.type_());
                return switch (self) {
                    .Some => |value| with.call(value),
                    .None => Return.none(),
                };
            }
        };
    }
};

fn double(v: u32) u32 {
    return v * 2;
}

fn option_test() void {
    const Option = Optional.Option;
    const opt = Option(u32).some(5);
    const b = opt.map(u32, double);

    std.debug.print("{}\n", .{b});

    const c = b.map_with(struct {
        val: i32,

        fn call(self: @This(), v: u32) i32 {
            return @intCast(i32, v) + self.val;
        }
    }{ .val = 3 });

    std.debug.print("{}\n", .{c});

    const d1: ?u32 = 4;
    const d = Optional.into_option(d1).map_with(struct {
        val: i32,

        fn call(self: @This(), v: u32) i32 {
            return @intCast(i32, v) + self.val;
        }
    }{ .val = 10 });
    std.debug.print("{}\n", .{d});

    std.debug.assert(b.Some == 10);
}

test "option" {
    std.debug.print("\n", .{});
    option_test();
}

const BTree = struct {
    const Self = @This();

    const B: usize = 4;
    // guaranteed to be odd, so can't insert into the middle of a full node
    // which would complicate splitting
    const CAPACITY: usize = 2 * B - 1;
    const MIN_AFTER_SPLIT: usize = B - 1;
    const E_LEFT_OF_CENTER: usize = B - 1;
    const E_RIGHT_OF_CENTER: usize = B;
    const NUM_EDGES: usize = 2 * B;

    ally: std.mem.Allocator = std.heap.c_allocator,
    root: ?NodeOrLeaf,

    fn create(ally: std.mem.Allocator) Self {
        return Self{
            .ally = ally,
            .root = null,
        };
    }

    fn insert(self: *Self, value: u32) !void {
        if (self.root) |*root| {
            const search = root.find_key(value);

            switch (search) {
                .Leaf => |_| {
                    return error.Occupied;
                },
                .Edge => |edge| {
                    const result = try edge.leaf.insert_value(value);
                    if (result) |split| {
                        // create new node which will replace self.
                        const parent = try Node.create(self.ally);
                        parent.leaf.level = split.left.level + 1;
                        NodeOrLeaf.from_leaf(parent.as_leaf()).push_value(split.middle);
                        parent.insert_node(NodeOrLeaf.from_leaf(split.left));
                        parent.insert_node(NodeOrLeaf.from_leaf(split.right));

                        self.root = .{ .internal = parent };
                    }
                },
            }
        } else {
            var leaf: *Leaf = try self.ally.create(Leaf);
            errdefer self.ally.destroy(leaf);
            leaf.init(self.ally);
            NodeOrLeaf.from_leaf(leaf).push_value(value);
            self.root = NodeOrLeaf{ .leaf = leaf };
        }
    }

    fn remove(self: *Self, key: u32) void {
        if (self.root) |*root| {
            const search = root.find_key(key);

            switch (search) {
                .Leaf => |leaf| {
                    NodeOrLeaf.from_leaf(leaf.leaf).remove_key_from_node(key);
                },
                .Edge => {},
            }
        }
    }

    fn find_key(self: *Self, key: u32) ?u32 {
        switch (self.root.?.find_key(key)) {
            .Leaf => |leaf| {
                return leaf.leaf.get_values()[leaf.idx];
            },
            else => {
                return null;
            },
        }
    }

    fn dbg(self: *Self) void {
        if (self.root) |root| {
            root.dbg();
            std.debug.print("\n", .{});
        }
    }

    fn destroy(self: *Self) void {
        if (self.root) |*root| {
            root.destroy();
        }
    }

    const NodeOrLeafTag = enum {
        internal,
        leaf,
    };

    const NodeOrLeaf = union(NodeOrLeafTag) {
        internal: *Node,
        leaf: *Leaf,

        fn force(self: NodeOrLeaf) NodeOrLeaf {
            return NodeOrLeaf.from_leaf(self.as_leaf());
        }

        fn destroy(self: NodeOrLeaf) void {
            switch (self.force()) {
                .internal => |node| {
                    node.destroy();
                },
                .leaf => |leaf| {
                    leaf.destroy();
                },
            }
        }

        fn as_leaf(self: NodeOrLeaf) *Leaf {
            switch (self) {
                .internal => |node| {
                    return node.as_leaf();
                },
                .leaf => |leaf| {
                    return leaf;
                },
            }
        }

        fn from_leaf(leaf: *Leaf) NodeOrLeaf {
            if (leaf.level == 0) {
                return .{ .leaf = leaf };
            } else {
                return .{ .internal = @ptrCast(*Node, leaf) };
            }
        }

        fn split_at(self: NodeOrLeaf, value: u32) !Leaf.SplitResult {
            const leaf = self.as_leaf();

            var idx: u16 = 0;
            for (leaf.get_values(), 0..) |v, i| {
                idx = @intCast(u16, i);
                if (v > value) {
                    break;
                }
            }

            std.debug.assert(leaf.len == CAPACITY);

            var new: *Leaf = undefined;
            switch (self) {
                .internal => |internal| {
                    const node = try Node.create(leaf.ally);
                    std.mem.copy(?NodeOrLeaf, &node.edges, internal.edges[B..]);

                    for (node.edges, 0..) |edge, i| {
                        if (edge) |edge_node| {
                            edge_node.as_leaf().set_parent(node, @intCast(u16, i));
                        }
                    }

                    new = node.as_leaf();
                },
                .leaf => {
                    const node = try Leaf.create(leaf.ally);
                    new = node;
                },
            }
            new.level = leaf.level;

            new.len = B - 1;
            std.mem.copy(u32, &new.values, leaf.values[B..]);
            const middle = leaf.values[B - 1];
            leaf.len = B - 1;

            // take from right half
            if (idx >= B) {
                NodeOrLeaf.from_leaf(new).push_value(value);
            } else {
                NodeOrLeaf.from_leaf(leaf).push_value(value);
            }

            return .{ .left = leaf, .middle = middle, .right = new };
        }

        fn push_value(self: NodeOrLeaf, value: u32) void {
            const leaf = self.as_leaf();

            std.debug.assert(leaf.len < CAPACITY);

            var n = leaf.len;
            for (leaf.get_values(), 0..) |val, i| {
                if (val >= value) {
                    n = @intCast(u16, i);
                    break;
                }
            }

            var tmp = value;
            for (leaf.get_values()[n..]) |*val| {
                const t = val.*;
                val.* = tmp;
                tmp = t;
            }
            leaf.values[leaf.len] = tmp;

            switch (self) {
                .internal => |node| {
                    var tmp2: ?NodeOrLeaf = null;
                    for (node.get_edges()[n + 1 ..]) |*edge| {
                        const t = edge.*;
                        edge.* = tmp2;
                        tmp2 = t;
                    }
                    node.edges[leaf.len + 1] = tmp2;
                },
                else => {},
            }

            leaf.len = leaf.len + 1;
        }

        fn remove_key_from_node(self: NodeOrLeaf, key: u32) void {
            switch (self.force()) {
                .internal => |node| {
                    //const Option = Optional.Option;
                    // need to check if the edges to the left and right of the to-be-removed key exist,
                    // and if both exist, find the immediate neighbors of the key and promote
                    // the one in the bigger leaf up as a replacement.

                    // this node will not require rebalancement in that case, but the leaf
                    // from which we removed the replacement key might.

                    // safety: we already know the key is in this exact node.
                    const index = node.leaf.find_key(key).Leaf.idx;

                    const left = Optional.into_option(node.get_edge(index));
                    const right = Optional.into_option(node.get_edge(index + 1));

                    if (right.is_some() and left.is_some()) {
                        const right_value = right.Some.find_least_significant_key();
                        const left_value = left.Some.find_most_significant_key();

                        // pick the leaf with more values to reduce unneccesairy rebalances
                        const myleaf = blk: {
                            if (right_value.leaf.len > left_value.leaf.len) {
                                // use right leaf
                                break :blk right_value;
                            } else {
                                // use left leaf
                                break :blk left_value;
                            }
                        };

                        const promoted = myleaf.leaf.get_values()[myleaf.idx];
                        myleaf.leaf.remove_key(promoted);
                        self.as_leaf().values[index] = promoted;

                        // TODO: check `myleaf` for rebalance
                        myleaf.leaf.rebalance();
                    } else {
                        // no adjacent edges
                        // (I don't think this can actually happen in a proper rtree)

                        // no need to promote any value from an adjecent subtree because
                        // there is no subtree on one side.

                        // instead we need to check for rebalance of self here TODO

                        node.shift_edges_left(index, 1);
                        self.as_leaf().rebalance();
                    }
                },
                .leaf => |node| {
                    // we can just remove the key and then check afterwards if the
                    // tree needs to be rebalanced

                    node.remove_key(key);

                    self.as_leaf().rebalance();
                },
            }
        }

        fn find_key(self: NodeOrLeaf, key: u32) Leaf.SearchResult {
            var leaf = self.as_leaf();
            while (true) {
                const search = leaf.find_key(key);

                switch (search) {
                    .Leaf => {
                        return search;
                    },
                    .Edge => |edge| {
                        const node = NodeOrLeaf.from_leaf(edge.leaf);
                        switch (node) {
                            .internal => |internal| {
                                if (internal.get_edges()[edge.idx]) |child| {
                                    leaf = child.as_leaf();
                                    // TODO: incredibly hacky I think..
                                    // gotta figure out WHERE this would even happen..
                                    // leaf.parent = .{ .parent = internal, .idx = edge.idx };
                                    // it happened when splitting nodes, and not reparenting
                                    // the split-off edges :|
                                    continue;
                                }
                            },
                            .leaf => {},
                        }
                        return search;
                    },
                }
            }
        }

        const SignificantKeyResult = struct {
            leaf: *Leaf,
            idx: u16,
        };

        fn find_most_significant_key(self: NodeOrLeaf) SignificantKeyResult {
            var node = self;
            while (true) {
                switch (node.force()) {
                    .internal => |internal| {
                        if (internal.get_edges()[internal.leaf.len]) |child| {
                            node = child;
                            continue;
                        }
                    },
                    .leaf => |leaf| {
                        return .{ .leaf = leaf, .idx = leaf.len - 1 };
                    },
                }
            }
        }

        fn find_least_significant_key(self: NodeOrLeaf) SignificantKeyResult {
            var node = self;
            while (true) {
                switch (node.force()) {
                    .internal => |internal| {
                        if (internal.get_edges()[0]) |child| {
                            node = child;
                            continue;
                        }
                    },
                    .leaf => |leaf| {
                        return .{ .leaf = leaf, .idx = 0 };
                    },
                }
            }
        }

        fn get_most_significant_value(self: NodeOrLeaf) u32 {
            switch (self.force()) {
                .internal => |node| {
                    if (node.get_edges()[node.leaf.len]) |edge| {
                        return edge.get_most_significant_value();
                    } else {
                        return node.as_leaf().get_values()[node.leaf.len - 1];
                    }
                },
                .leaf => |node| {
                    return node.get_values()[node.len - 1];
                },
            }
        }

        fn get_least_significant_value(self: NodeOrLeaf) u32 {
            switch (self.force()) {
                .internal => |node| {
                    if (node.get_edges()[0]) |edge| {
                        return edge.get_least_significant_value();
                    } else {
                        return node.as_leaf().get_values()[0];
                    }
                },
                .leaf => |node| {
                    return node.get_values()[0];
                },
            }
        }

        fn dbg(self: NodeOrLeaf) void {
            switch (self) {
                .internal => |node| {
                    node.dbg();
                },
                .leaf => |node| {
                    node.dbg();
                },
            }
        }
    };

    const Node = struct {
        leaf: Leaf,

        edges: [NUM_EDGES]?NodeOrLeaf = [_]?NodeOrLeaf{null} ** NUM_EDGES,

        fn create(ally: std.mem.Allocator) !*Node {
            var node = try ally.create(Node);
            node.init(ally);

            return node;
        }

        fn destroy(self: *Node) void {
            for (self.get_edges()) |edge| {
                if (edge) |*edg| {
                    edg.destroy();
                }
            }
            self.leaf.ally.destroy(self);
        }

        fn init(self: *Node, ally: std.mem.Allocator) void {
            self.* = Node{ .leaf = Leaf{ .ally = ally } };
        }

        fn as_leaf(self: *Node) *Leaf {
            return &self.leaf;
        }

        fn insert_node(self: *Node, child: NodeOrLeaf) void {
            const self_leaf = self.as_leaf();
            const ls = child.as_leaf().get_values()[0];

            var idx: u16 = self_leaf.len;
            for (self_leaf.get_values(), 0..) |v, i| {
                if (v > ls) {
                    idx = @intCast(u16, i);
                    break;
                }
            }

            if (self.get_edges()[idx]) |edge| {
                std.debug.print("edge already present:", .{});
                child.dbg();
                std.debug.print(" - ", .{});
                edge.dbg();
                std.debug.print("\n", .{});
            } else {
                child.as_leaf().set_parent(self, idx);
                self.get_edges()[idx] = child;
            }
        }

        const InsertResultTag = enum {
            Split,
            RightFromParent,
        };

        const InsertResult = union(InsertResultTag) {
            Split: Leaf.SplitResult,
            RightFromParent: *Leaf,
        };

        fn insert_split(self: *Node, split: Leaf.SplitResult) !?Leaf.SplitResult {
            const leaf = self.as_leaf();
            const value = split.middle;

            if (leaf.len < CAPACITY) {
                NodeOrLeaf.from_leaf(leaf).push_value(value);
                self.insert_node(NodeOrLeaf.from_leaf(split.right));
            } else {
                const parent_split = try leaf.split_at(value);
                const next_split = Leaf.SplitResult.concat(parent_split, split);

                if (leaf.parent) |parent| {
                    return parent.parent.insert_split(next_split);
                } else {
                    return next_split;
                }
            }

            return null;
        }

        fn get_edges(self: *Node) []?NodeOrLeaf {
            const len = self.leaf.len + 1;
            return self.edges[0..len];
        }

        fn get_edge(self: *Node, i: u16) ?NodeOrLeaf {
            if (i <= self.leaf.len) {
                const edge = self.get_edges()[i];
                if (edge) |node| {
                    return node.force();
                }
            }
            return null;
        }

        fn remove_edge(self: *Node, index: u16) ?NodeOrLeaf {
            const edge = self.get_edge(index);
            self.shift_edges_left(index + 1, 1);

            return edge;
        }

        fn shift_edges_left(self: *Node, start: u16, count: u16) void {
            std.debug.assert(start >= count);
            if (start <= self.leaf.len) {
                for (self.get_edges()[start..], start..) |edg, i| {
                    if (edg) |edge| {
                        edge.as_leaf().set_parent(self, @intCast(u16, i - count));
                        self.edges[i - count] = edge;
                    }
                }
            }
        }

        fn shift_edges_right(self: *Node, start: u16, count: u16) void {
            std.debug.assert(self.leaf.len - start >= count - 1);
            if (start <= self.leaf.len) {
                // shift edges right
                for (count..(self.leaf.len + 1 - start)) |i| {
                    const idx = self.leaf.len - 1 - count - i;
                    if (self.get_edge(@intCast(u16, idx))) |edge| {
                        edge.as_leaf().set_parent(self, @intCast(u16, idx + count));
                        self.edges[idx + count] = edge;
                        // set moved eges to null
                        self.edges[idx] = null;
                    }
                }
            }
        }

        fn dbg(self: *Node) void {
            const values = self.leaf.get_values();
            const edges = self.get_edges()[0..values.len];
            std.debug.print("{{ ", .{});
            std.debug.print("[{}] ", .{self.leaf.level});

            // an internal node should have at least B elements
            if (self.leaf.len < MIN_AFTER_SPLIT) {
                std.debug.print("[E] ", .{});
            }

            for (values, edges) |v, e| {
                if (e) |edge| {
                    edge.dbg();
                    std.debug.print(", ", .{});
                }
                std.debug.print("{}, ", .{v});
            }
            if (self.get_edges()[values.len]) |edge| {
                edge.dbg();
            }
            std.debug.print(" }}", .{});
        }
    };

    const ParentPtr = struct {
        parent: *Node,
        idx: u16,

        fn into_node_or_leaf(self: ?ParentPtr) ?NodeOrLeaf {
            if (self) |ptr| {
                return .{ .node = ptr.parent };
            } else {
                return null;
            }
        }
    };

    const Leaf = struct {
        ally: std.mem.Allocator,
        level: usize = 0,
        parent: ?ParentPtr = null,

        len: u16 = 0,
        values: [CAPACITY]u32 = undefined,

        fn create(ally: std.mem.Allocator) !*Leaf {
            var leaf = try ally.create(Leaf);
            leaf.init(ally);

            return leaf;
        }

        fn dbg(self: *Leaf) void {
            const values = self.get_values();
            std.debug.print("{any}", .{values});
        }

        fn init(self: *Leaf, ally: std.mem.Allocator) void {
            self.* = Leaf{ .ally = ally };
        }

        fn destroy(self: *Leaf) void {
            self.ally.destroy(self);
        }

        fn set_parent(self: *Leaf, parent: *Node, idx: u16) void {
            if (idx < 0 or idx > parent.leaf.len or idx > CAPACITY) {
                std.debug.print("{}", .{idx});
            }

            self.parent = .{ .parent = parent, .idx = idx };
        }

        const SplitResult = struct {
            // attached, old node that may be modified
            left: *Leaf,
            // lose value, previously attacked, must be inserted
            // if we go all the way to the top, the root node must have some value and this is it
            middle: u32,
            // new, free floating leaf, must be attached
            right: *Leaf,

            fn dbg(self: SplitResult) void {
                std.debug.print("[ ", .{});
                self.left.dbg();
                std.debug.print(", {}, ", .{self.middle});
                self.right.dbg();
                std.debug.print(" ]", .{});
            }

            fn dbg_verbose(self: SplitResult) void {
                std.debug.print("[ ", .{});
                NodeOrLeaf.from_leaf(self.left).dbg();
                std.debug.print(", {}, ", .{self.middle});
                NodeOrLeaf.from_leaf(self.right).dbg();
                std.debug.print(" ]", .{});
            }

            fn concat(parent: SplitResult, child: SplitResult) SplitResult {
                // safety @ptrCast(): we know parent left and right are nodes because
                // they originated from childs parent

                // we only care about the childs middle and left, and since they are ordered we
                // can learn about the right part from the middle part
                if (child.middle < parent.middle) {
                    // I'm not sure if checking the childs right part is actually needed?
                    // I don't think so but can't think of a solid enough reason why

                    // child is entirely between two values of the parent, so any relation between
                    // the childs mid point and any of the parents values is true for any of
                    // the childs values, right?
                    @ptrCast(*Node, parent.left).insert_node(NodeOrLeaf.from_leaf(child.right));
                }
                // since they cant be equal, this must mean child is bigger than parent
                else {
                    @ptrCast(*Node, parent.right).insert_node(NodeOrLeaf.from_leaf(child.right));
                }

                return parent;
            }
        };

        fn split_at(self: *Leaf, value: u32) !SplitResult {
            return NodeOrLeaf.from_leaf(self).split_at(value);
        }

        const SearchResultTag = enum { Edge, Leaf };

        const SearchResult = union(SearchResultTag) {
            Edge: struct { leaf: *Leaf, idx: u16 },
            Leaf: struct { leaf: *Leaf, idx: u16 },
        };

        fn needs_rebalance(self: *Leaf) bool {
            return self.len < MIN_AFTER_SPLIT;
        }

        fn rebalance(self: *Leaf) void {
            if (self.needs_rebalance()) {
                if (self.parent) |parent| {
                    const nonnull = blk: {
                        if (parent.idx == 0) {
                            break :blk Optional.Option(u16).none();
                        } else {
                            break :blk Optional.Option(u16).some(parent.idx);
                        }
                    };
                    const left_sib = nonnull.join_with(struct {
                        node: *Node,
                        fn call(a: @This(), idx: u16) Optional.Option(NodeOrLeaf) {
                            return Optional.into_option(a.node.get_edge(idx - 1));
                        }
                    }{ .node = parent.parent });

                    const right_sib = Optional.into_option(parent.parent.get_edge(parent.idx + 1));

                    const Side = enum { Left, Right };

                    const sibling: struct { node: NodeOrLeaf, side: Side } = blk: {
                        switch (left_sib) {
                            .Some => |left| {
                                switch (right_sib) {
                                    .Some => |right| if (left.as_leaf().len > right.as_leaf().len) {
                                        break :blk .{ .node = left, .side = .Left };
                                    } else {
                                        break :blk .{ .node = right, .side = .Right };
                                    },
                                    .None => {
                                        break :blk .{ .node = left, .side = .Left };
                                    },
                                }
                            },
                            .None => {
                                switch (right_sib) {
                                    .Some => |right| {
                                        break :blk .{ .node = right, .side = .Right };
                                    },
                                    .None => {
                                        // unreachable
                                    },
                                }
                            },
                        }
                    };

                    const offset = @intCast(i32, sibling.node.as_leaf().parent.?.idx) - @intCast(i32, parent.idx);
                    std.debug.print("{any} {} {} {}\n", .{ sibling.node.as_leaf().parent, sibling.node.as_leaf().parent.?.idx, parent.idx, offset });

                    if (sibling.node.as_leaf().len > MIN_AFTER_SPLIT) {
                        // rotate
                        const index = @intCast(usize, @max(0, parent.idx + offset));
                        std.debug.print("{} {} {}\n", .{ parent.idx, offset, index });

                        NodeOrLeaf.from_leaf(self).push_value(parent.parent.leaf.values[index]);
                        parent.parent.leaf.values[index] = blk: {
                            if (offset < 0) {
                                const value = sibling.node.get_most_significant_value();
                                sibling.node.as_leaf().remove_key(value);
                                break :blk value;
                            } else {
                                const value = sibling.node.get_least_significant_value();
                                sibling.node.as_leaf().remove_key(value);
                                break :blk value;
                            }
                        };

                        // TODO: rotate edges
                        switch (NodeOrLeaf.from_leaf(self)) {
                            .internal => |iself| {
                                const isibling = sibling.node.internal;
                                iself.shift_edges_right(0, 1);
                                iself.edges[0] = isibling.edges[isibling.leaf.len];
                                isibling.edges[isibling.leaf.len] = null;

                                iself.edges[0].?.as_leaf().set_parent(iself, 0);
                            },
                            .leaf => {},
                        }
                    } else {
                        // merge
                        const left = switch (sibling.side) {
                            .Left => sibling.node.as_leaf(),
                            .Right => self,
                        };
                        const right = switch (sibling.side) {
                            .Right => sibling.node.as_leaf(),
                            .Left => self,
                        };

                        defer right.destroy();

                        // merge parent seperator into left
                        const key = parent.parent.leaf.get_value(left.parent.?.idx);
                        parent.parent.leaf.remove_key(key);
                        NodeOrLeaf.from_leaf(left).push_value(key);

                        // remove superfluous parent edge to right
                        _ = parent.parent.remove_edge(right.parent.?.idx);

                        // copy values from right to left
                        std.mem.copy(u32, left.values[left.len..], right.values[0..right.len]);

                        // TODO: copy edges from right to left aswell
                        switch (NodeOrLeaf.from_leaf(left)) {
                            .internal => |ileft| {
                                const iright = @ptrCast(*Node, right);

                                std.mem.copy(?NodeOrLeaf, ileft.edges[0..], iright.edges[0..]);

                                for (ileft.get_edges(), 0..) |e, i| {
                                    if (e) |ee| {
                                        ee.as_leaf().set_parent(ileft, @intCast(u16, i));
                                    }
                                }
                            },
                            .leaf => {},
                        }

                        left.len = left.len + right.len;

                        parent.parent.as_leaf().rebalance();
                    }
                } else {
                    // root node: if len is 0, just copy paste from child node if exists
                    if (self.len == 0) {
                        switch (NodeOrLeaf.from_leaf(self)) {
                            .internal => |root| {
                                if (root.edges[0]) |edge| {
                                    defer edge.destroy();
                                    // copy values
                                    root.as_leaf().len = edge.as_leaf().len;
                                    root.as_leaf().level = edge.as_leaf().level;
                                    std.mem.copy(u32, root.leaf.values[0..], edge.as_leaf().values[0..]);

                                    switch (edge.force()) {
                                        .internal => |child| {
                                            std.mem.copy(?NodeOrLeaf, root.edges[0..], child.edges[0..]);

                                            for (root.get_edges(), 0..) |e, i| {
                                                if (e) |ee| {
                                                    ee.as_leaf().set_parent(root, @intCast(u16, i));
                                                }
                                            }
                                        },
                                        .leaf => {
                                            // idk. just empty?
                                            for (&root.edges) |*e| {
                                                e.* = null;
                                            }
                                        },
                                    }
                                }
                            },
                            .leaf => {
                                // idk. just empty?
                            },
                        }
                    }
                }
            }
        }

        fn remove_key(self: *Leaf, key: u32) void {
            // TODO: remove keys from internal nodes
            // safety: key must be contained by .values
            var n: u16 = 0;
            for (self.get_values(), 0..) |val, i| {
                if (val >= key) {
                    n = @intCast(u16, i);
                    break;
                }
            }

            for (self.get_values()[n + 1 .. self.len], n..) |val, j| {
                self.values[j] = val;
            }
            self.len = self.len - 1;
        }

        fn find_key(self: *Leaf, key: u32) SearchResult {
            for (self.get_values(), 0..) |v, i| {
                if (key < v) {
                    // decending left of v
                    return .{ .Edge = .{ .leaf = self, .idx = @intCast(u16, i) } };
                } else if (key == v) {
                    // located key at v
                    return .{ .Leaf = .{ .leaf = self, .idx = @intCast(u16, i) } };
                }
            }

            // decending right-most
            return .{ .Edge = .{ .leaf = self, .idx = self.len } };
        }

        // returns null on success, or a split result which could not be merged
        // up because we are at the root node
        fn insert_value(self: *Leaf, value: u32) !?SplitResult {
            const leaf = self;

            if (leaf.len < CAPACITY) {
                NodeOrLeaf.from_leaf(leaf).push_value(value);
            } else {
                var split = try leaf.split_at(value);

                if (leaf.parent) |parent| {
                    return parent.parent.insert_split(split);
                } else {
                    return split;
                }
            }

            return null;
        }

        fn get_values(self: *Leaf) []u32 {
            const len = self.len;
            return self.values[0..len];
        }

        fn get_value(self: *Leaf, idx: u16) u32 {
            return self.get_values()[idx];
        }
    };
};

pub fn main() !void {
    // Prints to stderr (it's a shortcut based on `std.io.getStdErr()`)
    std.debug.print("All your {s} are belong to us.\n", .{"codebase"});

    // stdout is for the actual output of your application, for example if you
    // are implementing gzip, then only the compressed bytes should be sent to
    // stdout, not any debugging messages.
    const stdout_file = std.io.getStdOut().writer();
    var bw = std.io.bufferedWriter(stdout_file);
    const stdout = bw.writer();

    try stdout.print("Run `zig build test` to run the tests.\n", .{});

    try bw.flush(); // don't forget to flush!
}

fn printValues(leaf: *BTree.Leaf) void {
    const values = leaf.get_values();
    std.debug.print("{any}\n", .{values});
}

// test "leaf split" {
//     std.debug.print("testing splitting\n", .{});

//     var tree = BTree.create(std.testing.allocator);
//     defer tree.destroy();
//     try tree.insert(2);
//     try tree.insert(4);
//     try tree.insert(6);
//     try tree.insert(3);
//     try tree.insert(7);
//     std.debug.print("before split:", .{});
//     printValues(tree.root.?.as_leaf());

//     const split = try tree.root.?.as_leaf().split_at(5);

//     std.debug.print("after split:", .{});
//     printValues(tree.root.?.as_leaf());

//     std.debug.print("split: {?}\n", .{split});
//     tree.ally.destroy(split.right);
// }

// test "btree insert" {
//     std.debug.print("testing insertion\n", .{});
//     var tree = BTree.create(std.testing.allocator);
//     defer tree.destroy();
//     try tree.insert(10);
//     try tree.insert(4);
//     try tree.insert(6);
//     try tree.insert(3);
//     try tree.insert(9);
//     try tree.insert(8);
//     tree.dbg();
// }

test "btree seq insert" {
    std.debug.print("sequential insertions\n", .{});

    var tree = BTree.create(std.testing.allocator);
    defer tree.destroy();

    for (0..100) |i| {
        tree.insert(@intCast(u32, i)) catch {
            std.debug.print("{} already present - ignoring\n", .{i});
        };
    }
    tree.dbg();
}

test "btree rand insert" {
    std.debug.print("random insertions\n", .{});

    var tree = BTree.create(std.testing.allocator);
    defer tree.destroy();
    var buf = std.ArrayList(u32).init(std.testing.allocator);
    defer buf.deinit();

    var rng = std.rand.DefaultPrng.init(0);

    for (0..10000) |_| {
        // const i = rng.random().intRangeAtMost(u32, 0, 512);
        const i = rng.random().int(u32);
        try buf.append(i);
        tree.insert(i) catch {
            std.debug.print("{} already present - ignoring\n", .{i});
        };
    }

    for (buf.items) |i| {
        if (tree.find_key(i)) |_| {} else {
            std.debug.print("{} lost\n", .{i});
        }
    }

    tree.dbg();
}

test "ref removal" {
    std.debug.print("remove keys\n", .{});

    var tree = BTree.create(std.testing.allocator);
    defer tree.destroy();

    for (0..20) |i| {
        _ = try tree.insert(@intCast(u32, i));
    }

    tree.dbg();
    std.debug.print("\n", .{});

    for (0..15) |i| {
        tree.remove(@intCast(u32, i));
    }

    tree.dbg();
    std.debug.print("\n", .{});
}

test "btree rebalance" {
    std.debug.print("rebalance simple leaves\n", .{});

    var tree = BTree.create(std.testing.allocator);
    defer tree.destroy();

    for (0..9) |i| {
        _ = try tree.insert(@intCast(u32, i));
    }

    tree.dbg();
    std.debug.print("\n", .{});

    {
        const node = tree.root.?.internal.edges[1].?;
        node.as_leaf().len = node.as_leaf().len - 1;
        node.as_leaf().rebalance();

        tree.dbg();
        std.debug.print("\n", .{});
    }

    {
        const node = tree.root.?.internal.edges[1].?;
        node.as_leaf().len = node.as_leaf().len - 1;
        node.as_leaf().rebalance();

        tree.dbg();
        std.debug.print("\n", .{});
    }

    {
        const node = tree.root.?.internal.edges[1].?;
        node.as_leaf().len = node.as_leaf().len - 1;
        node.as_leaf().rebalance();

        tree.dbg();
        std.debug.print("\n", .{});
    }
}

test "btree least most sig value" {
    std.debug.print("least/most significant values\n", .{});

    var tree = BTree.create(std.testing.allocator);
    defer tree.destroy();

    for (0..100) |i| {
        tree.insert(@intCast(u32, i)) catch {
            std.debug.print("{} already present - ignoring\n", .{i});
        };
    }
    tree.dbg();
    std.debug.print("\n", .{});

    {
        const l = tree.root.?.get_least_significant_value();
        const r = tree.root.?.get_most_significant_value();
        std.debug.print("{} < {}\n", .{ l, r });
        std.debug.assert(l == 0);
        std.debug.assert(r == 99);
    }

    {
        const node = tree.root.?;
        const l = node.find_least_significant_key();
        const r = node.find_most_significant_key();
        std.debug.print("{} < {}\n", .{ l, r });
    }

    {
        const node = tree.root.?.internal.edges[1].?.force();
        const l = node.get_least_significant_value();
        const r = node.get_most_significant_value();
        std.debug.print("{} < {}\n", .{ l, r });
    }
}
